// 16-thread-stacks.c, 64-Bit-Version
// 
// Systemprogrammierung, SS 2013
// Hans-Georg Esser

#include <pthread.h>
#include <stdio.h>    // printf, snprintf
#include <unistd.h>   // sleep, alarm
#include <stdlib.h>   // exit, system
#include <signal.h>   // signal (), SIGALRM

// Makro-Definitionen
#define get_esp(v) __asm__ ("mov %%rsp, %0" : "=r"(v))
#define get_ebp(v) __asm__ ("mov %%rbp, %0" : "=r"(v))
#define PEEK(addr) (*(unsigned long*)(addr))

#define print_stack(name,adr,count) {  \
  int i;                               \
  for (i=0; i<count; i++) {            \
    printf (#name ": [%012lx] = %012lx\n", \
      adr+i*8, PEEK(adr+i*8));         \
    fflush (stdout);                   \
  }                                    \
}

#define print_regs(name) \
  printf (#name ": esp = 0x%012lx, ebp = 0x%012lx, eip = 0x%012lx\n", \
    esp, ebp, eip);

#define show_mem_map(s,pid) \
  snprintf (s, 100, "cat /proc/%d/maps", pid); \
  printf ("%s\n", s); fflush (stdout); system (s);

void get_eip (unsigned long *v) {
  unsigned long eip;
  __asm__ __volatile__ ("movq 8(%%rbp), %0" : "=r" (eip));
  // __asm__ __volatile__ ("popq %%rax; movq %%rax, %0; pushq %%rax" : "=r" (eip));
  *v = eip;
}

#define get_regs get_esp (esp); get_ebp (ebp); get_eip (&eip)

void *f1 () {
  unsigned long esp, ebp, eip;
  get_regs;
  print_regs (f1__);
  print_stack (f1__, esp, 3);
  sleep (2);
  return NULL;
}

void *f2 () {
  unsigned long esp, ebp, eip;
  get_regs;
  print_regs (f2__);
  sleep (2);
  return NULL;
}

void alarm_handler (int sig) {
  unsigned long esp, ebp, eip;
  printf ("alrm: sig = %d\n", sig);

  // alarm: Register
  get_regs;
  print_regs (alrm);
  print_stack (alrm, esp, 3);
  return;
}

int main () {
  unsigned long esp, ebp, eip;
  char s[100];  // for command string
  int pid = getpid ();
  printf ("pid = %d\n", pid);

  signal (SIGALRM, alarm_handler);  // alarm handler
  alarm (1);                        // alarm nach 1 sek.

  // Memory Map anzeigen
  show_mem_map (s, pid);
  
  // Threads starten
  pthread_t thr1, thr2, thr3;
  pthread_create (&thr1, NULL, f1, NULL);
  pthread_create (&thr2, NULL, f2, NULL);
  pthread_create (&thr3, NULL, f2, NULL);
  
  // main: Register
  get_regs;
  print_regs (main);
  print_stack (main, esp, 3);

  // Memory Map anzeigen
  show_mem_map (s, pid);
  
  // Threads einsammeln
  pthread_join (thr1, NULL);
  pthread_join (thr2, NULL);
  pthread_join (thr3, NULL);

  // Fertig  
  exit(0);
}
